<?php

namespace Ulmus\Ldap\Query;

use Ulmus\Common\{ EntityField, Sql };

use Ulmus\Query\{ QueryBuilderInterface, Fragment };

class Filter extends Fragment {
    const OPERATOR_LIKE = "LIKE";
    const OPERATOR_EQUAL = "=";
    const OPERATOR_NOT_EQUAL = "<>";
    const CONDITION_AND = "&";
    const CONDITION_OR  = "|";
    const CONDITION_NOT = "!";
    const COMPARISON_IN = "IN";
    const COMPARISON_IS = "IS";
    const COMPARISON_NULL = "NULL";

    public array $conditionList;

    public QueryBuilderInterface $queryBuilder;

    public ? Filter $parent = null;

    public string $condition = self::CONDITION_AND;

    public function __construct(? QueryBuilderInterface $queryBuilder, $condition = self::CONDITION_AND)
    {
        $this->queryBuilder = $queryBuilder;
        $this->condition = $condition;
        $this->parent = $queryBuilder->Filter ?? null;
    }
    
    public function add($field, $value, string $operator, string $condition, bool $not = false) : self
    {
        $this->conditionList[] = [
            $field,
            $value,
            $operator ?: $this->queryBuilder->conditionOperator,
            $condition,
            $not
        ];

        return $this;
    }

    public function render(bool $skipToken = false) : array
    {
        $stack = [];
        
        foreach ($this->conditionList ?? [] as $key => $item) {
            if ( $item instanceof Filter ) {
                if ( $item->conditionList ?? false ) {
                    $stack[] = ( $key !== 0 ? "{$item->condition} " : "" ) . "(" . implode('', $item->render($skipToken)) . ")";
                }
            }
            else {
                list($field, $value, $operator, $condition, $not) = $item;
                $stack[] = $latest = $this->FilterCondition($field, $value, $operator, $key !== 0 ? $condition : "", $not);
            }
        }

        return [
            'filters' => implode('', $stack)
        ];
    }

    protected function FilterCondition($field, $value, string $operator = self::OPERATOR_EQUAL, string $condition = self::CONDITION_AND, bool $not = false) {
        return new class($this->queryBuilder, $field, $value, $operator, $condition, $not) {

            public $value;
            public bool $not = false;
            public string $field;
            public string $operator;
            public string $condition;
            public QueryBuilderInterface $queryBuilder;

            protected string $content = "";

            public function __construct(QueryBuilderInterface $queryBuilder, string $field, $value, string $operator, string $condition, bool $not) {
                $this->queryBuilder = $queryBuilder;
                $this->field = $field;
                $this->value = $value;
                $this->condition = $condition;
                $this->operator = $operator;
                $this->not = $not;
            }

            public function render() : string
            {
                $value = $this->value();

                return $this->content ?: $this->content = implode("", array_filter([
                    "(",
                    $this->condition,
                    $this->not ? Filter::CONDITION_NOT : "",
                    $this->field,
                    $this->operator(),
                    $value,
                    ")",
                ]));
            }

            protected function operator() : string
            {
                if ( is_array($this->value) ) {
                    return (in_array($this->operator, [ '!=', '<>' ]) ? Filter::CONDITION_NOT . " " : "") . Filter::COMPARISON_IN;
                }

                return $this->operator;
            }

            protected function value()
            {
                if ( is_array($this->value) ) {
                    $stack = [];

                    foreach($this->value as $item) {
                        $stack[] = $this->filterValue($item);
                    }

                    return "(" . implode(", ", $stack) . ")";
                }

                return $this->filterValue($this->value);
            }

            protected function filterValue($value)
            {
                if ( $value === null ) {
                    throw new \Exception("NULL is not an acceptable value for filter request.");
                }
                elseif ( is_object($value) && ( $value instanceof EntityField ) ) {
                    return $value->name();
                }
                else {
                    return $value;
                }
            }

            public function __toString() : string
            {
                return $this->render();
            }
        };
    }
}
